import logging
from time import time

from bubuku.aws import AWSResources
from bubuku.aws.cluster_config import ClusterConfig
from bubuku.aws.ec2_node_launcher import Ec2NodeLauncher
from bubuku.aws.node import Ec2Node
from bubuku.broker import BrokerManager
from bubuku.controller import Change
from bubuku.zookeeper import BukuExhibitor

_LOG = logging.getLogger('bubuku.features.rolling_restart')


class RollingRestartChange(Change):
    def __init__(self, zk: BukuExhibitor, cluster_config: ClusterConfig,
                 restart_assignment,
                 broker_id: str,
                 image: str,
                 instance_type: str,
                 scalyr_key: str,
                 scalyr_region: str,
                 kms_key_id: str):
        self.zk = zk
        self.restart_assignment = restart_assignment
        self.broker_id = broker_id
        self.broker_id_to_restart = self.restart_assignment.pop(broker_id)
        self.broker_ip_to_restart = self.zk.get_broker_address(self.broker_id_to_restart)

        self.cluster_config = cluster_config
        self.cluster_config.set_application_version(image)
        self.cluster_config.set_instance_type(instance_type)
        self.cluster_config.set_scalyr_account_key(scalyr_key)
        self.cluster_config.set_scalyr_region(scalyr_region)
        self.cluster_config.set_kms_key_id(kms_key_id)

        self.aws = AWSResources(region=self.cluster_config.get_aws_region())
        self.ec_node = Ec2Node(self.aws, self.cluster_config, self.broker_ip_to_restart)
        self.ec2_node_launcher = Ec2NodeLauncher(self.aws, self.cluster_config)
        self.cluster_config.set_availability_zone(self.ec_node.get_node_availability_zone())

        self.state_context = StateContext(self.zk, self.aws, self.ec_node, self.ec2_node_launcher,
                                          self.broker_id_to_restart, self.restart_assignment)

    def get_name(self) -> str:
        return 'rolling_restart'

    def can_run(self, current_actions):
        return all([a not in current_actions for a in ['stop', 'restart', 'rebalance']])

    def run(self, current_actions) -> bool:
        if not self._is_cluster_in_good_state():
            _LOG.error('Cluster is not stable skipping restart iteration')
            return True

        return self.state_context.run()

    def _is_cluster_in_good_state(self):
        return True


class StartBrokerChange(Change):
    def __init__(self, zk: BukuExhibitor, broker: BrokerManager):
        self.zk = zk
        self.broker = broker

    def get_name(self):
        return 'start'

    def can_run(self, current_actions):
        return all([a not in current_actions for a in ['restart', 'stop']])

    def run(self, current_actions):
        zk_conn_str = self.zk.get_conn_str()
        try:
            self.broker.start_kafka_process(zk_conn_str)
        except Exception as e:
            _LOG.error('Failed to start kafka process against {}'.format(zk_conn_str), exc_info=e)
            return True
        return False


class StateContext:
    def __init__(self, zk: BukuExhibitor, aws: AWSResources, ec_node: Ec2Node, ec2_node_launcher: Ec2NodeLauncher,
                 broker_id_to_restart, restart_assignment):
        self.zk = zk
        self.restart_assignment = restart_assignment
        self.aws = aws
        self.ec_node = ec_node
        self.ec2_node_launcher = ec2_node_launcher
        self.broker_id_to_restart = broker_id_to_restart
        self.current_state = StopKafka(self)

    def run(self):
        """
        Runs states one after another. If state is finished, it takes the next one.
        """
        try:
            _LOG.info('Running state {}'.format(self.current_state))
            if self.current_state.run():
                next_state = self.current_state.next()
                _LOG.info('Next state {}'.format(next_state))
                if next_state is None:
                    return False
                self.current_state = next_state
            return True
        except Exception as e:
            _LOG.error('Failed to run state', exc_info=e)
            return True


class State:
    """
    State which can be run as many times as required before it finishes it work. The progress of the state has to be
    recoverable
    """

    def __init__(self, state_context):
        self.state_context = state_context
        self.time_to_check_s = time()

    def run(self) -> bool:
        """
        Runs the state, and if state finishes successfully it returns True, otherwise it returns False, which means
        that state has to be executed again
        """
        pass

    def next(self):
        """
        Return the next state, which has to be executed after the current state
        """
        pass

    def run_with_timeout(self, func):
        """
        Runs func() with timeout
        :param func function to execute
        :param timeout_s timeout before executing state next time
        """
        if time() >= self.time_to_check_s:
            self.time_to_check_s = time() + 10
            return func()
        return False


class StopKafka(State):
    def run(self):
        from bubuku.features.remote_exec import RemoteCommandExecutorCheck
        RemoteCommandExecutorCheck.register_stop(self.state_context.zk, self.state_context.broker_id_to_restart)
        return True

    def next(self):
        return WaitBrokerStopped(self.state_context)

    def __str__(self):
        return 'StopKafka: stopping broker {}'.format(self.state_context.broker_id_to_restart)


class WaitBrokerStopped(State):
    def run(self):
        def func():
            return not self.state_context.zk.is_broker_registered(self.state_context.broker_id_to_restart)

        return self.run_with_timeout(func)

    def next(self):
        return DetachVolume(self.state_context)

    def __str__(self):
        return 'WaitBrokerStopped: waiting for broker {} to stop'.format(self.state_context.broker_id_to_restart)


class DetachVolume(State):
    def run(self):
        self.state_context.ec_node.detach_volume()
        return True

    def next(self):
        return TerminateInstance(self.state_context)

    def __str__(self):
        return 'DetachVolume: detaching volume {} from broker {}'.format(self.state_context.ec_node.get_volume_id(),
                                                                         self.state_context.broker_id_to_restart)


class TerminateInstance(State):
    def run(self):
        self.state_context.ec_node.terminate()
        return True

    def next(self):
        return WaitVolumeAvailable(self.state_context)

    def __str__(self):
        return 'WaitBrokerStopped: waiting for broker {} to stop'.format(self.state_context.broker_id_to_restart)


class WaitVolumeAvailable(State):
    def run(self):
        def func():
            return self.state_context.ec_node.is_volume_available()

        return self.run_with_timeout(func)

    def next(self):
        return LaunchInstance(self.state_context)

    def __str__(self):
        return 'WaitVolumeAvailable: waiting for volume {} to be available'.format(
            self.state_context.ec_node.get_volume_id())


class LaunchInstance(State):
    def run(self):
        self.state_context.ec2_node_launcher.launch()
        return True

    def next(self):
        return WaitVolumeAttached(self.state_context)


class WaitVolumeAttached(State):
    def run(self):
        def func():
            return self.state_context.ec_node.is_volume_in_use()

        return self.run_with_timeout(func)

    def next(self):
        return StartKafka(self.state_context)

    def __str__(self):
        return 'WaitVolumeAttached: waiting for volume {} to be attached'.format(
            self.state_context.ec_node.get_volume_id())


class StartKafka(State):
    def run(self):
        from bubuku.features.remote_exec import RemoteCommandExecutorCheck
        RemoteCommandExecutorCheck.register_start(self.state_context.zk, self.state_context.broker_id_to_restart)
        return True

    def next(self):
        return WaitKafkaRunning(self.state_context)

    def __str__(self):
        return 'StartKafka: starting broker {}'.format(self.state_context.broker_id_to_restart)


class WaitKafkaRunning(State):
    def run(self):
        def func():
            return self.state_context.zk.is_broker_registered(self.state_context.broker_id_to_restart)

        return self.run_with_timeout(func)

    def next(self):
        return RegisterRollingRestart(self.state_context)

    def __str__(self):
        return 'WaitKafkaRunning: waiting broker {} is running'.format(self.state_context.broker_id_to_restart)


class RegisterRollingRestart(State):
    def run(self):
        if len(self.state_context.restart_assignment) == 0:
            _LOG.info('Rolling restart is successfully finished')
        else:
            action = {'name': 'rolling_restart',
                      'restart_assignment': self.state_context.restart_assignment,
                      'image': self.state_context.cluster_config.get_application_version(),
                      'instance_type': self.state_context.cluster_config.get_instance_type(),
                      'scalyr_key': self.state_context.cluster_config.get_scalyr_region(),
                      'scalyr_region': self.state_context.cluster_config.get_scalyr_account_key(),
                      'kms_key_id': self.state_context.cluster_config.get_kms_key_id()}
            next_broker_id = list(self.state_context.restart_assignment.keys())[0]
            self.state_context.zk.register_action(action, broker_id=next_broker_id)
        return True

    def next(self):
        return None



import logging

from bubuku.aws import AWSResources
from bubuku.aws.cluster_config import ClusterConfig

_LOG = logging.getLogger('bubuku.aws.node')

KAFKA_LOGS_EBS = 'kafka-logs-ebs'


class Ec2Node(object):
    def __init__(self, aws: AWSResources, cluster_config: ClusterConfig, ip: str):
        self.aws = aws
        self.cluster_config = cluster_config
        self.ip = ip
        self.instance = self._get_instance_by_ip()
        _LOG.info('Searching for instance %s volumes', self.instance.instance_id)
        volumes = self.aws.ec2_client.describe_instance_attribute(InstanceId=self.instance.instance_id,
                                                                  Attribute='blockDeviceMapping')
        data_volume = next(v for v in volumes['BlockDeviceMappings'] if v['DeviceName'] == '/dev/xvdk')
        data_volume_id = data_volume['Ebs']['VolumeId']
        self.volume = self.aws.ec2_resource.Volume(data_volume_id)

    def get_node_availability_zone(self):
        return self.volume.availability_zone

    def get_volume_id(self):
        return self.volume.id

    def get_ip(self):
        return self.ip

    def is_volume_in_use(self):
        self.volume.load()
        if self.volume.state == 'in-use':
            _LOG.info('Volume %s is attached. Clearing tag:Name', self.volume)
            self.volume.create_tags(Tags=[{'Key': 'Name', 'Value': ''}])
            _LOG.info('Completed clearing tag:Name for %s', self.volume)
            return True
        return False

    def is_volume_available(self):
        self.volume.load()
        return self.volume.state == 'available'

    def detach_volume(self):
        self.volume.create_tags(Tags=[{'Key': 'Name', 'Value': KAFKA_LOGS_EBS}])
        _LOG.info('Detaching %s from %s', self.volume.id, self.instance.instance_id)
        self.aws.ec2_client.detach_volume(VolumeId=self.volume.id, Force=False)

    def terminate(self):
        cluster_name = self.cluster_config.get_cluster_name()
        _LOG.info('Terminating %s in %s', self.instance, cluster_name)
        alarm_name = '{}-{}-auto-recover'.format(cluster_name, self.instance.instance_id)
        _LOG.info('Deleting alarm %s in %s for %s', alarm_name, cluster_name, self.instance.instance_id)
        self.aws.cloudwatch_client.delete_alarms(AlarmNames=[alarm_name])
        self.instance.terminate()

    def is_terminated(self):
        self.instance.load()
        _LOG.info('Instance state is %s. Waiting ...', self.instance.state['Name'])
        if self.instance.state['Name'] == 'terminated':
            _LOG.info('%s is successfully terminated', self.instance)
            return True
        return False

    def _get_instance_by_ip(self):
        instances = self.aws.ec2_resource.instances.filter(Filters=[
            {'Name': 'instance-state-name', 'Values': ['running', 'pending']},
            {'Name': 'network-interface.addresses.private-ip-address', 'Values': [self.ip]},
            {'Name': 'tag:Name', 'Values': [self.cluster_config.get_cluster_name()]}])
        instances = list(instances)
        if not instances:
            raise Exception('Instance by ip {} not found in cluster {}'
                            .format(self.ip, self.cluster_config.get_cluster_name()))
        _LOG.info('Found %s by ip %s', instances[0], self.ip)
        return instances[0]



import logging

import netaddr

from bubuku.aws import AWSResources
from bubuku.aws.cluster_config import ClusterConfig

_LOG = logging.getLogger('bubuku.cluster.aws.subnet')


class IpAddressAllocator(object):
    def __init__(self, aws: AWSResources, cluster_config: ClusterConfig):
        self.aws = aws
        self.cluster_config = cluster_config

    def _get_subnets(self, prefix_filter: str) -> list:
        """
        Returns lists of subnets, which names start
        with the specified prefix (it should be either 'dmz-' or
        'internal-'), sorted by the Availability Zone and filtered by vpc id
        """
        vpc_id = self.cluster_config.get_vpc_id()
        availability_zone = self.cluster_config.get_availability_zone()
        _LOG.info('Getting subnets for vpc_id: %s and availability_zone: %s', vpc_id, availability_zone)

        resp = self.aws.ec2_client.describe_subnets()
        subnets = []

        for subnet in resp['Subnets']:
            if subnet['VpcId'] != vpc_id:
                continue
            for tag in subnet['Tags']:
                if tag['Key'] == 'Name' and tag['Value'].startswith(prefix_filter):
                    if availability_zone:
                        if subnet['AvailabilityZone'] == availability_zone:
                            subnets.append(subnet)
                    else:
                        subnets.append(subnet)
        _LOG.info('Got subnets %s ', subnets)
        return subnets

    def allocate_ip_addresses(self, address_count: int) -> list:
        """
        Allocate unused private IP addresses by checking the current
        reservations
        Return list of tuples (subnet, ip)
        """
        _LOG.info('Allocating IP addresses ...')

        def try_next_address(ips, subnet):
            try:
                return str(next(ips))
            except StopIteration:
                raise Exception('Out of available IP addresses in subnet {}'.format(subnet['CidrBlock']))

        #
        # Here we have to account for the behavior of launch_*_nodes
        # which iterate through subnets to put the instances into
        # different Availability Zones.
        #
        subnets = self._get_subnets('internal-')
        network_ips = [netaddr.IPNetwork(s['CidrBlock']).iter_hosts() for s in subnets]

        for idx, ips in enumerate(network_ips):
            #
            # Some of the first addresses in each subnet are
            # taken by AWS system instances that we can't see,
            # so we try to skip them.
            #
            for _ in range(10):
                try_next_address(ips, subnets[idx])

        i = 0
        result_ips_subnets = []
        while i < address_count:
            idx = i % len(subnets)
            subnet = subnets[idx]
            ip = try_next_address(network_ips[idx], subnet)
            resp = self.aws.ec2_client.describe_instances(Filters=[{
                'Name': 'private-ip-address',
                'Values': [ip]
            }])
            if not resp['Reservations']:
                i += 1
                _LOG.info('Got ip address %s ', ip)
                result_ips_subnets.append((subnet, ip))

        _LOG.info('IP Addresses are allocated')

        return result_ips_subnets
